#!/usr/bin/perl -w

# BridgeMats.pl - Copyright 2008 Pete Matthews

$VERSION = "BridgeMats Version A802";

# Version  Date         By   Description
# A704      1-Apr-2007  PDM  New - not for BAM (yet).
# A704a     5-Apr-2007  PDM  Bug fixes and improvements for Individual Movements
# A704b     7-Apr-2007  PDM  Switch file (BridgeMats.sw)
# A704c    14-Apr-2007  PDM  Fixes for scale factor, arrow switch
# A704d    22-Apr-2007  PDM  Arrow switch detection for single table (H0712)
# A705     10-May-2007  PDM  Add -w, -L.
# A705a    20-May-2007  PDM  Add Orange & Brown.  First published version
# A709      3-Sep-2007  PDM  Baron/Barclay sleeves -f 1.11.  Old ACBL -f 1.15.
# A710      6-Oct-2007  PDM  Trivial help/switch changes, posting PDFs.
# A802     24-Feb-2008  PDM  Trivial help/switch changes.

sub Usage() {
#        1         2         3         4         5         6         7
#234567890123456789012345678901234567890123456789012345678901234567890123456789
  print <<"EndUsage";

                           $VERSION Copyright (c) 2008

Usage:  BridgeMats.pl  [switches]  movement-file 

where movement-file is an ACBLscore movement file and switches are:

  -b number  number of boards per round (default for ~24 boards)
  -c number  text color (0 in BridgeMats.sw):
             0=use number of tables, black for partial tables
             1=Black (program default), 2=Olive, 3=Purple, 
             4=Green, 5=Blue, 6=Teal, 7=Red, 8=Orange, 9=Brown.
  -d         dump the movement file as text to stdout (see -n)
             Use Notepad to print after redirecting (> file.txt)
  -f factor  scale factor: default 1.0 for US Letter.  Use 1.27 for full
             US Legal, 1.11 for Baron/Barclay sleeves, 1.15 for old ACBL.
  -g number  group this number of ACBLscore rounds as one - when
             applicable, -g is required for correct next table
  -h         print this help to stdout and exit
  -l         omit the last mat (dummy table in ACBLscore)
  -L         print the license for BridgeMats and exit
  -n         produce no PostScript mats, dump only (sets -d)
  -o number  produce only the mat for this table number
  -r number  roving pair number (value from file is ignored)
  -t number  truncate movement after this number of grouped rounds -
             27 ungrouped rounds is the maximum without overprinting
  -w         double-weave movement (even EW down, odd EW up), board jump

BridgeMats.pl writes a PostScript file of guide cards (table mats)
from a binary ACBLscore duplicate bridge movement file.
You should run it in a DOS box.  Next, open the resulting file
in GSview, where you may examine, print and/or convert to a PDF.
The PostScript file could also be tweaked using Notepad - its name
is the same as the movement file, but with the .PS extension.

For best results, the current working directory should contain 
the movement file (whose extension must be .MOV, .BAM, or .IND),
BridgeMats.pl, BridgeMats.ps, and BridgeMats.sw.  Movements
normally come from C:\\ACBLSCOR\\MOV, \\BAM or \\IND - or
wherever you saved them from ACBLscore.

Perl must be installed on your system and in your path.  Get it for
MS Windows from http://www.activestate.com/Products/ActivePerl/.
Get GSview and Ghostscript from http://www.ghostgum.com.au/
or http://www.cs.wisc.edu/~ghost/gsview/ - both are needed.

You'll probably need to run BridgeMats.pl, to determine the best
switches to use for a new movement.  Some movements are difficult
or impossible to diagnose.  Check the results!

A full command line example:

    bridgemats.pl -c 0 -d -g 2 -l -r 13 H7ROVER.MOV

BridgeMats.sw provides specific defaults for the tested cases below.
Unless otherwise noted, the movements are ACBLscore "External".

  A full command line example (others rely on BridgeMats.sw):
    13 pairs:    bridgemats.pl -c 0 -d -g 2 -l -r 13 H7ROVER.MOV

  One-winner movements, generally 24 boards:
     4 pairs:    bridgemats.pl H0203.MOV   -- save from ACBLscore
     4 pairs:    bridgemats.pl H0206.MOV   -- ACBLscore external
     6 pairs:    bridgemats.pl H0305.MOV   -- save from ACBLscore
     8 pairs:    bridgemats.pl H0406.MOV   -- save from ACBLscore
    10 pairs:    bridgemats.pl H0508.MOV   -- save from ACBLscore
    12 pairs:    bridgemats.pl H0608.MOV   -- save from ACBLscore
    13 pairs:    bridgemats.pl HAJ7RVR.MOV -- from Harold Jordan
    14 pairs:    bridgemats.pl H0712.MOV   -- save from ACBLscore

  12-board movements for Bridge Plus:
     6 pairs:    bridgemats.pl H0305B+.MOV -- save from ACBLscore
     8 pairs:    bridgemats.pl H0406B+.MOV -- save from ACBLscore
    10 pairs:    bridgemats.pl H0506B+.MOV -- save from ACBLscore

  Double-weave movements:
    16 pairs:    bridgemats.pl M0808.MOV
    24 pairs:    bridgemats.pl M1212DW.MOV -- from Tim Francis-Wright

  Individual movements, with rotation, when possible:
     8 players:  bridgemats.pl S0207.IND
    10 players:  bridgemats.pl S0310A.IND
    12 players:  bridgemats.pl S0311A.IND
    13 players:  bridgemats.pl S0413.IND
    14 players:  bridgemats.pl S0414.IND
    16 players:  bridgemats.pl S0412A.IND
    18 players:  bridgemats.pl S0509.IND
    20 players:  bridgemats.pl S0515.IND
    22 players:  bridgemats.pl S0611.IND
  * 24 players:  bridgemats.pl S0621.IND
    26 players:  bridgemats.pl S0713M.IND
    28 players:  bridgemats.pl S0721.IND
  * 32 players:  bridgemats.pl S0824.IND
    36 players:  bridgemats.pl S0924.IND
  * 40 players:  bridgemats.pl S1027.IND
    44 players:  bridgemats.pl S1133.IND
  * 48 players:  bridgemats.pl S1224.IND
    52 players:  bridgemats.pl S1326.IND

  * Personal guide cards are required for this movement.

EndUsage
#        1         2         3         4         5         6         7
#234567890123456789012345678901234567890123456789012345678901234567890123456789
  exit 1;
  }

sub License() {
  # The MIT License from http://www.opensource.org/licenses/mit-license.php
  print <<"EndLicense";

Copyright (c) 2008 Peter D. Matthews, Jr.

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

EndLicense
#        1         2         3         4         5         6         7
#234567890123456789012345678901234567890123456789012345678901234567890123456789
  exit 1;
  }
# -- Declarations

use strict 'subs';
sub ReadSwitchFile();
sub ProcessSwitches($@);
sub FindNext($$$);
sub ReadMovementFile();
sub PrintSetupVars();
sub SetNext($$);
sub PrintPageVars();

#----------------------------------------------------------------#

sub ReadSwitchFile() {

  my $i;
  my $j;
  my $c;
  open (SWFILE, $SWFILE) or 
      warn "--> unable to open switch file: $SWFILE <--";
  $nfsw = 0;
  my $swstate = 0;

  while (<SWFILE>) {
    $c = substr($_,0,1);
    ($c eq '#') && next;     # comment
    chomp; # remove newline
    my $l = length($_);
    my $v;
    if ($l > 2) { $v = substr($_,2); }
    else        { $v = ""; }
    if ($c eq '%') {         # -- movement file section
      $v =~ tr/a-z \t/A-Z/d; # uppercase without whitespace
      ($swstate==2) && last; # all done at end of local section
      if ( "$v" eq "$FILE" ) { 
        $swstate = 2;        # this is our local section
        next;
        }
      else { $swstate = 1; } # wrong local section
      }
    ($swstate==1) && next;   # switch in wrong local section
    # switch is global or in correct local section
    if ($c ge '0' and $c le '9' ) {
      $j = ord($c) - ord('0');
      $desc[$c] = $v;
      $DI[$c] = 1;
      }
    else { # add to file switch array
      $fsw[$nfsw] = "-$c";
      $nfsw++;
      if ($l > 2) {
        $fsw[$nfsw] = $v;
        $nfsw++;
        }
      }
    }
  close(SWFILE);

  print "Defaults:";
  for ($i=0; $i<$nfsw; $i++) { print " $fsw[$i]"; }
  print "\n";
  }

#----------------------------------------------------------------#

sub ProcessSwitches($@) {

  my ($source,@Switch) = @_;

  while ($#Switch >= 0) {
    if ($Switch[0] eq "-b") {
      shift @Switch;
      $BoardsPerRound = $Switch[0];
      shift @Switch;
      if ( $BoardsPerRound < 1 or $BoardsPerRound > 36 )
        { die "--> not between 1 and 36: -b $BoardsPerRound <--"; }
      }
    elsif ($Switch[0] eq "-c") {
      shift @Switch;
      $NColor = $Switch[0];
      shift @Switch;
      if ( $NColor < 0 or $NColor > $MaxColor )
          { die "--> not between 0 and $MaxColor: -c $NColor <--"; }
      }
    elsif ($Switch[0] eq "-d") {
      shift @Switch;
      $DUMP = 1;
      }
    elsif ($Switch[0] eq "-f") {
      shift @Switch;
      $FACT = $Switch[0];
      shift @Switch;
      if ( $FACT < .5 or $FACT > 2 ) 
          { die "--> not between .5 and 2: -f $FACT <--"; }
      }
    elsif ($Switch[0] eq "-g") {
      shift @Switch;
      $GROUP = $Switch[0];
      shift @Switch;
      if ( $GROUP < 1 or $GROUP >5 )
          { die "--> not between 1 and 5: -g $GROUP <--"; }
      }
    elsif ($Switch[0] eq "-h") {
      Usage;
      }
    elsif ($Switch[0] eq "-l") {
      shift @Switch;
      $PrintLast = 0;
      }
    elsif ($Switch[0] eq "-L") {
      License;
      }
    elsif ($Switch[0] eq "-n") {
      shift @Switch;
      $PrintMats = 0;
      $DUMP = 1;
      }
    elsif ($Switch[0] eq "-o") {
      shift @Switch;
      $PrintOnly = $Switch[0];
      shift @Switch;
      if ( $PrintOnly<1 or $PrintOnly>40 ) 
          { die "--> not between 1 and 40: -o $PrintOnly <--"; }
      }
    elsif ($Switch[0] eq "-r") {
      shift @Switch;
      $Rover = $Switch[0];
      shift @Switch;
      if ( $Rover<1 or $Rover>80 ) 
          { die "--> not between 1 and 80: -r $Rover <--"; }
      }
    elsif ($Switch[0] eq "-t") {
      shift @Switch;
      $TruncAfter = $Switch[0];
      shift @Switch;
      if ( $TruncAfter<2 or $TruncAfter>36 ) 
          { die "--> not between 2 and 36: -t $TruncAfter <--"; }
      }
    elsif ($Switch[0] eq "-w") {
      $Weave = 1;
      shift @Switch;
      }
    else {
      my $msg;
      if ($source==1)  {
        $msg = "switch file";
        warn 
    "--> if previous switch has no value, check line ends before column 3 <--";
        }
      else {
        $msg = "command line";
        }
      die "--> Unrecognized $msg switch: $Switch[0] <--";
      }
    }
  }

#----------------------------------------------------------------#

sub ReadMovementFile() {
  open (MOV, $file) or die "--> unable to open switch defaults file: $file <--\n";
  binmode (MOV); # otherwise read converts 0x0d0a (CRLF) to 0x0a (LF) - H7ROVER

  my ($buf, @byte);
  my ($c1, $c2, $c3, $c4);
  my $signature;
  my $tablehdr;
  my $breaktest;
  my $NB;

  # -- read the first 2 bytes of the header, convert to hexadecimal
  #    Note: unable to get ActivePerl to convert 2 bytes to signed integer)
  read (MOV, $buf, 1);
  $c1 = unpack('H',$buf);
  $c2 = unpack('h',$buf);
  read (MOV, $buf, 1);
  $c3 = unpack('H',$buf);
  $c4 = unpack('h',$buf);
  $signature = "${c1}${c2}${c3}${c4}";

  if    ( $signature eq "5d8d" ) {                                # = -19347
    if ( $TYPE eq "BAM" )        { $movement = "Board-a-Match"; }
    else                         { $movement = "Pair"; }
    }
  elsif ( $signature eq "55ba" ) { $movement = "Individual"; }    # = -27835
  elsif ( $signature eq "6db4" ) { $movement = "Board-a-Match"; } # (B0807TH.BAM)
  else { 
  die "--> signature=$signature - not an ACBLscore movement file: $file <--\n";
    }
  ($DUMP) && print "Type:     ",  $movement, "    (", $signature, ")\n";
  # -- now read the next four header bytes
  read (MOV, $buf, 4);
  @byte = unpack ('c*', $buf);
  $howell = $byte[0];
  $tables = $byte[1];
  $pairs = $tables*2;   # adjust later
  $players = $tables*4;
  $rounds = $byte[2];
  $boards = $byte[3];
  ($DUMP) && print "Howell:   $howell\n";
  ($DUMP) && print "Tables:   $tables\n";
  ($DUMP) && print "Rounds:   $rounds\n";
  ($DUMP) && print "Boards:   $boards      (if only 1 board/round)\n";

  # -- Set up table numbers - use indexes except in TN and TNP arrays
  for ($h=0; $h<$tables; $h++) {
    $t = $h +1;
    $TN[$h] = $t;       # table number (numeric)
    if    ($t == 6)       { $TNP[$h] = "SIX"; }
    elsif ($t == 9)       { $TNP[$h] = "NINE"; }
    else                  { $TNP[$h] = $t; } 
    }

  # -- read description
  ($DUMP) && print 
  "Description:       1         2         3         4         5         6\n";
  ($DUMP) && print 
  "          123456789012345678901234567890123456789012345678901234567890\n";
  for ($t=0; $t<10; $t++) {
    read (MOV, $buf, 60);
    ($DUMP) && print "      $t=> ", $buf, "\n";
    if ($DI[$t]) {
      ($DUMP) && print "    ++$t=> ", $desc[$t], "\n";
      }
    else {
      $desc[$t] = $buf;
      }
    }

  # -- read rest of header
  if ( $TYPE eq "IND" ) {
    ($DUMP) && print 
    "    [no pair/board/phantom/rover info for Individual movement]\n";
    @board = (1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,
        19,20,21,22,23,24,25,26,27,28,29,30,31,32,33,34,35,36);

    # -- Set up for Rest of Read
    $ND = 4;  # number of directions
    $b = 4;   # index of boards in cell array
    $NB = 5;  # number of bytes to read
    $n = 5;   # index of notes in cell array
    $breaktest = 3;
    $tablehdr = ":  (N,E,S,W,BD) --------------------------------------\n";
    }

  else {      #  MOV or BAM
    read (MOV, $buf, 1);
    $highpair = ord($buf);
    ($DUMP) && print "HighPair: $highpair\n";
    read (MOV, $buf, 64);
    @pair = unpack ('c*', $buf);
    ($DUMP) && print "Pairs:    ";
    for ($t=0; $t<64; $t++) {
      ($DUMP) && print $pair[$t], " ";
      if ( ($t%16) == 15 ) { 
        ($DUMP) && print "\n";
        if ($t < 63) {print "          "; }
        }
      }
    $phantom = $pair[61];
    if ($phantom>0) {
      ($DUMP) && print "Phantom:  $phantom NS\n";
      }
    elsif ($phantom<0) {
      ($DUMP) && print "Phantom:  $phantom EW\n";
      }
    else {
      ($DUMP) && print "Phantom:  $phantom\n";
      }
    $rover = $pair[63];
    if ($rover>0) {
      ($DUMP) && print "Rover:    EW\n";
      }
    elsif ($rover<0) {
      ($DUMP) && print "Rover:    NS\n";
      }
    else {
      ($DUMP) && print "Rover:    $rover\n";
      }

    # -- read board data     
    read (MOV, $buf, 36);
    @board = unpack ('c*', $buf);
    ($DUMP) && print "Boards:   ";
    for ($t=0; $t<36; $t++) { 
      ($DUMP) && print $board[$t], " ";
      if ( $t == 17 ) { ($DUMP) && print "\n          "; }
      }
    ($DUMP) && print "\n";

    if ( $highpair != 0 ) { die "--> unable to handle renumbered pairs at this time<--"; }

    # -- Set up for Rest of Read
    $ND = 2;  # number of directions
    $b = 2;   # index of boards in cell array
    $NB = 3;  # number of bytes to read
    $n = 3;   # index of notes in cell array
    $breaktest = 5;
    $tablehdr = ":  (NS,EW,BD) ----------------------------------------\n";
    }

  # -- Read & Check Table Records
  for ($t=0; $t<$tables; $t++) {
    ($DUMP) && print "Table ", $t+1, $tablehdr;
    for ($r=0; $r<$rounds; $r++) {
      read (MOV, $buf, $NB) or die "--> bad read - file is probably damaged <--";
      @byte = unpack ('c*', $buf) or die "--> bad unpack <--";
      ($DUMP) && print "(";
      for ($d=0; $d<=$ND; $d++) {  # read boards as well as directions
        $cell[$d][$r][$t] = $byte[$d];
        ($DUMP) && print $byte[$d];
        if ( $d < $ND ) { ($DUMP) && print ","; }
        }
      ($DUMP) && print ") ";
      if ( $r%($breaktest+1) == $breaktest ) { ($DUMP) && print "\n"; }
      # initialize Notes
      $cell[$n][$r][$t] = "";
      }
    ($DUMP) && print "\n";
    # initialize for the checking loop
    for ($d=0; $d<$ND; $d++) {
      $NextTable[$d][$t] = -1; # index, -2=unpredictable
      $NextDir[$d][$t] = -1;   # direction or -2=switch
      }
    }
  close(MOV);

  # -- Check tables, groups, rounds
  if ( $PrintOnly>$tables ) 
      { die "--> exceeds number of tables: -o $PrintOnly <--"; }
  if ( $GROUP > 1 ) {
    $bpr = $GROUP * int($rounds/$GROUP);
    ($bpr=$GROUP) ||
      die "--> not a factor of number of $rounds ($$rounds): -g $GROUP <--";
    $bpr = $GROUP * int($boards/$GROUP);
    ($bpr=$GROUP) ||
      die "--> not a factor of number of boards ($boards): -g $GROUP <--";
    }
  my $T = $TruncAfter * $GROUP;
  if ( $T < $rounds ) {
    $rounds = $T;
    ($DUMP) && print "Rounds:   $rounds    (-t)\n";
    }

  if ($BoardsPerRound == 0) {
    $BoardsPerRound = int(29/$rounds);
    $bpr="BridgeMats default, see -b";
    }
  else { $bpr="from -b"; }
  $BoardsTotal = $rounds * $BoardsPerRound;
  ($DUMP) && print "Bd/Rd:    $BoardsPerRound       ($bpr)\n";
  ($DUMP) && print "Played:   $BoardsTotal      (calculated)\n";

  # -- Check for Relays, Initialize
  for ($t=0; $t<$tables; $t++) {
    $RealTable[$t] = 0;
    $SwitchPair[$t] = -1;
    for ($r=0; $r<$rounds; $r++) { # check for fake tables
      $bd = $cell[$b][$r][$t];  
      if ($bd>0) { $RealTable[$t] = 1; }
      }
    for ($r=0; $r<$rounds; $r+=$GROUP) { # by grouped rounds
      $bd = $cell[$b][$r][$t];  
      $set = 0;
      # -- check for relays
      for ($x=0; $x<$tables; $x++) {
        ($x == $t) && next;  # skip current table
        for ($g=0; $g<$GROUP; $g++) {
          if ($cell[$b][$r+$g][$x] == $bd) {
            if ($set) {  # more than 2 table relay
              $cell[$n][$r][$t]    = "Relay";
              $cell[$n][$r+$g][$t] = "Relay";
              }
            else {       # assume 2-table relay for now
              $set = 1;
              my $xp = $x + 1;
              $cell[$n][$r][$t]    = "Relay T$xp";
              $cell[$n][$r+$g][$t] = "Relay T$xp";
              }
            }
          }
        }
      }
    }

  # -- if pairs, check for switch tables
  if ( $TYPE eq "MOV" ) { 
    for ($t=0; $t<$tables; $t++) {
      my $x = $cell[0][0][$t];
      my $y = $cell[1][0][$t];
      my $u = $cell[0][1][$t];
      my $v = $cell[1][1][$t];
      my $z = -1;
      my $d = -1;
      for ($r=$GROUP; $r<$rounds; $r+=$GROUP) {
        $u = $cell[0][$r][$t];
        $v = $cell[1][$r][$t];
        ($u==$Rover) && next;
        ($v==$Rover) && next;
        if ($z>-1) {  # test switch pair candidate
          if ($u==$z) {
            if ($d==-1)    { $d = 0; } 
            elsif ($d!=-2) {  
              if ($d==1)   { $d = -2; } # switch
              }
            next;
            }
          elsif ($v==$z) {
            if ($d==-1)    { $d = 1; } 
            elsif ($d!=-2) {  
              if ($d==0)   { $d = -2; } # switch
              }
            next;
            }
          else            { $z = -1; last; }  # not a switch table
          }
        else {  # look for switch pair candidate 
          if    ($u==$x or $v==$x) { $z = $x; }
          elsif ($u==$y or $v==$y) { $z = $y; }
          else                     { last; }  # not a switch table
          }
        }
      if ($z>-1 and $d==-2) { $SwitchPair[$t] = $z; }
      }
    }

  # -- find next tables
  for ($t=0; $t<$tables; $t++) {
    for ($r=0; $r<$rounds-$GROUP; $r+=$GROUP) { # omit last round
      for ($d=0; $d<$ND; $d++) {
        if ($NextTable[$d][$t]>=-1) { FindNext($d,$r,$t); }
        if ($SwitchPair[$t]>-1) { last; }  # only check one dir
        }
      }
    }

  # -- check Double-Weave movement
  if ($Weave == 1) {
    $Half = int($tables/2);
    if ( $tables != (2*$Half) ) {
      die "Number of tables must be even for double-weave: -w";
      }
    my $w;
    my $x;
    my $y;
    my $z;
    for ($t=0; $t<$tables; $t++) { 
      for ($r=0; $r<$rounds-$GROUP; $r+=$GROUP) {
        $x = $cell[1][$r][$t];
        $z= int($x/2);
        if ($x == (2*$z)) {
          if ($t>0)          { $y = $t - 1; }
          else               { $y = $tables - 1; }
          }
        elsif ($t<$tables-1) { $y = $t + 1; }
        else                 { $y = 0; }
        $w = $cell[1][$r+1][$y];
        if ($w != $x) {
          print "\nr=$r, t=$t   w=$w, x=$x, y=$y, z=$z \n";
          die "Expected EW movement not found for double-weave";
          }
        }
      }
    }
  }

#----------------------------------------------------------------#

sub FindNext($$$) {  # called once for each table/round/direction
  my($d,$r,$t) = @_;
  my $sp = $SwitchPair[$t];
  my $pr = $cell[$d][$r][$t];

  if ($r+$GROUP >= $rounds) { return; }  # can't test last round
  elsif ($SwitchPair[$t]>-1) {  # test the correct direction (only)
    if ($sp == $pr) {
      $d = 1;
      $pr = $cell[$d][$r][$t];
      $NextDir[$d][$t] = -2;  # don't check direction at target table
      }
    }
  if ($Rover == $pr) { return; }  # skip if this is rover
  my $n = $NextTable[$d][$t];
  my $j = $r + $GROUP;
  my $k = $NextDir[$d][$t];
  my $tg;
  my $i;
  my $l;
  if ($n == -2)      { return; }  # unpredictable - do nothing

  if ($n > -1) {
    if ($k > -1) { # check expected location
      $tg = $cell[$k][$j][$n];
      if ($pr==$tg or $Rover==$tg) { goto RETURN; }
      }
    for ($l=0; $l<$ND; $l++) { # check same table
      $tg = $cell[$l][$j][$n];
      if ($pr==$tg or $Rover==$tg) { # at same table
        $NextDir[$d][$t] = -2 ;
#        ($DUMP) &&
#          print "d=$d, r=$r, t=$t   k=$k, j=$j, n=$n, l=$l  pr=$pr tg=$tg",
#                " -- direction change\n";
        goto RETURN;
        }
      }
    $NextTable[$d][$t] = -2;  # not found
#    ($DUMP) &&
#      print "d=$d, r=$r, t=$t   k=$k, j=$j, n=$n",
#            " -- not found\n";
    goto RETURN;
    }
  else {
    # search entire next round
    for ($k=0; $k<$ND; $k++) {
      for ($i=0; $i<$tables; $i++) {
        if ($RealTable[$i]==0 and $Rover) { # skip fake table (H7ROVER)
#          ($DUMP) &&
#            print "d=$d, r=$r, t=$t   k=$k, j=$j, i=$i",
#                  " -- skip fake table with rover\n";
          next;
          }
        if ($pr==$cell[$k][$j][$i]) {
          $NextTable[$d][$t] = $i;
          $NextDir[$d][$t] = $k;
          ($SwitchPair[$t]>-1)  and $NextDir[$d][$t] = -2;
#          ($DUMP) &&
#            print "d=$d, r=$r, t=$t   k=$k, j=$j, i=$i, n=$n",
#                  " -- found next table\n";
          goto RETURN;
          }
        }       
      }
    # if not found, just check next time
#      ($DUMP) &&
#        print "d=$d, r=$r, t=$t   k=$k, j=$j, i=$i",
#              " -- next table not found yet\n";
    }
  return;

  RETURN:
  if ($SwitchPair[$t]>-1) { # brute force
    $n = $NextTable[$d][$t];
    $NextTable[0][$t] = $n;
    $NextTable[1][$t] = $n;
    $NextDir[0][$t] = -2;
    $NextDir[1][$t] = -2;
    }
  return;
  }  

#----------------------------------------------------------------#

sub PrintSetupVars() {
  my $t;
  my $r;
  my $d;
  my $p;
  my $test;
  # reduce number of pairs...
  if    ($TYPE eq "BAM") {
    $Title1 = "(Board-a-Match Bridge)";
    $Title2 = "(Movement for)";
    $PAIRS  = "($tables TEAMS)";
    }
  elsif ($TYPE eq "IND") {
    # -- deduct phantom players from total (S0713M)
    for ($t=0; $t<$tables; $t++) {
      if ( ! $RealTable[$t] ) {
        for ($d=0; $d<$ND; $d++) {
          $p = $cell[$d][0][$t];
          $test = 0;
          for ($r=1; $r<$rounds; $r++) {
            if ($p!=$cell[$d][$r][$t]) {
              $test = 1;
              next;
              }
            }
          $players -= $test;
          }
        }
      }
    $Title1 = "(Individual Bridge)";
    $Title2 = "(Movement for)";
    $PAIRS  = "($players PLAYERS)";
    $BoxLine[0] = "/BoxLine {";
    $BoxLine[1] = "  NT n get";
    $BoxLine[2] = "  BD n get";
    $BoxLine[3] = "  WW n get";
    $BoxLine[4] = "  SS n get";
    $BoxLine[5] = "  EE n get";
    $BoxLine[6] = "  NN n get";
    $BoxLine[7] = "  RD n get";
    $BoxLine[8] = "  IndBoxLine } def";
    $NBoxLine = 9;
    @DirArr = ("NN", "EE", "SS", "WW");
    $add=4;
    }
  else { #        "MOV"
    ($Rover>0 || $phantom!=0) && $pairs--; 
    $Title1 = "(Duplicate Bridge)";
    $Title2 = "(Movement for)";
    $PAIRS  = "($pairs PAIRS)";
    $BoxLine[0] = "/BoxLine {";
    $BoxLine[1] = "  NT n get";
    $BoxLine[2] = "  BD n get";
    $BoxLine[3] = "  EW n get";
    $BoxLine[4] = "  NS n get";
    $BoxLine[5] = "  RD n get";
    $BoxLine[6] = "  PairsBoxLine } def";
    $NBoxLine = 7;
    @DirArr = ("NS", "EW");
    $add=0;
    }
  # -- print "here document" with interpolation
  print PSOUT <<"PrintEndSetup";
/VERSION ($VERSION) def
/FACT    $FACT def
/TITLE   ($FILE) def
/Title1  $Title1 def   
/Title2  $Title2 def
/PAIRS   $PAIRS def
/ROUNDS  $rounds def
/DESC [
 ($desc[0])
 ($desc[1])
 ($desc[2])
 ($desc[3])
 ($desc[4])
 ($desc[5])
 ($desc[6])
 ($desc[7])
 ($desc[8])
 ($desc[9])
 ] def
PrintEndSetup

  for ($i=0; $i<$NBoxLine; $i++) {
    print PSOUT "$BoxLine[$i] \n";
    }
  }

#----------------------------------------------------------------#

sub SetNext($$) {

  my ($d, $t) = @_;
  my $k = $NextDir[$d][$t];
  my $n = $NextTable[$d][$t];
  my $Next;
  my $Other;

  if ($DUMP) {
    print "Next for T", $TN[$t], ", ", $Dir[$d+4];
    ($d%2==1) && print " ";
    print " ==> ";
#    print "(k=$k, n=$n)   ";
#    if (! $RealTable[$n]) { print "  Sit Out"; }
    if ($n > -1) { 
      print "T", $TN[$n];
      if ($k > -1) { print ", ", $Dir[$k+4]; }
      }
    if ($SwitchPair[$t]>-1) 
        { print " - Switch Table, Pair ", $SwitchPair[$t]; }
    print "\n";
    }

  if ($n<0) { # unpredictable
    $Next = "()";
    $Other = "()";
    return ($Next, $Other);
    }
  elsif ($n==$t) { # this table
    if ($k==$d) {
      if ($RealTable[$t]) { 
        if ($GROUP==1)    { $Next = "(Remain Stationary)"; }
        else              { $Next = "(Remain at This Table)"; }
        }
      else                { $Next = "(Phantom Player)"; }
      $Other = $Next;
      return ($Next, $Other);
      }
    # else drop through
    }

  if ($k>-1) { # specific table and direction
#    ($DUMP) && print "...      t=$t, d=$d      n=$n, k=$k\n";
    if (($ND==4) or ($k==$d)) {  # use second half of @Dir
      if ($SwitchPair[$t]==-1) { $k+=4; }
      }
    if ($SwitchPair[$t]>-1) 
         { $Next = "(Other Pairs Go Next to Table $TN[$n])"; }
    else { $Next = "(Go Next to Table $TN[$n], $Dir[$k])"; }
    if ($k%4 < 2) {
      if ($SwitchPair[$t]>-1) 
           { $Other = "(Other Pairs Go Next to Table $TN[$n])"; }
      else { $Other = "(Go Next to Table $TN[$n], $Dir[$k+2])"; }
      }
    else {
      $Other = "(--)"; # should never be used 
      }
    }
  else { # another table, direction switches
    if ($SwitchPair[$t]>-1) 
         { $Next = "(Other Pairs Go Next to Table $TN[$n])"; }
    else { $Next = "(Go Next to Table $TN[$n])"; }
    $Other = $Next;
    }
  return ($Next, $Other);
  }

#----------------------------------------------------------------#

sub PrintPageVars() {

  my $la1NS = 3;
  my $la1EW = 3;
  my $up;
  my $down;
  my $L2N = "()";
  my $L2S = "()";
  my $L2E = "()";
  my $L2W = "()";
  my $Note = "(Note)";

  if ($TYPE eq "IND") {
    ($NextNorth, $ignore) = SetNext(0,$t);
    ($NextEast,  $ignore) = SetNext(1,$t);
    ($NextSouth, $ignore) = SetNext(2,$t);
    ($NextWest,  $ignore) = SetNext(3,$t);
    }
  else {  # MOV or BAM
    # set North-South
    ($NextNorth, $NextSouth) = SetNext(0,$t);

    # set East-West
    if ($Weave) {
      if ($t>0)          { $down = $t - 1; }
      else               { $down = $tables - 1; }
      if ($t<$tables-1)  { $up = $t + 1; }
      else               { $up = 0; }
      $NextEast = "(EVEN-number pair DOWN to table $TN[$down])";
      $L2E = "(ODD-number pair UP to table $TN[$up])";
      $NextWest = $NextEast;
      $L2W = $L2E;
      $la1EW = -9;
      # set Notes for board movement
      $Note="(Move Bds)";
      my $b = 2;
      my $n = 3;
      my $r;
      my $s;
      my $bd;
      my $next;
      for ($r=0; $r<$rounds-$GROUP; $r+=$GROUP) { # by grouped rounds
        $bd = $cell[$b][$r][$t];
        $next = -1;
        for ($s=0; $s<$tables; $s++) {
          if ($bd == $cell[$b][$r+1][$s]) {
            $next = $s;
            last;
            }
          }
        ($next == -1 ) and die "Bad board layout in weave movement.";
        $cell[$n][$r][$t] = "to Table $TN[$next]";
        }
      }
    elsif ($SwitchPair[$t]>-1) { 
      $L2N = $NextNorth;
      $L2E = $L2N;
      $L2S = $L2N;
      $L2W = $L2N;
      $NextNorth = 
          "(Pair $SwitchPair[$t] Remain at This Table)";
      $NextEast = $NextNorth;
      $NextSouth = $NextNorth;
      $NextWest = $NextNorth;
      $la1NS = -9;
      $la1EW = -9;
      }
    else {
      ($NextEast, $NextWest)  = SetNext(1,$t);
      }
    }

  print PSOUT "$Color[$NColor]\n";

  # -- Rounds
  print PSOUT "/RD [ (Rd.)";
  $l=0;
  for ($r=0; $r<$rounds; $r++) {
    if ($r%$GROUP == 0) {
      $l++;
      print PSOUT " (${l}.)";
      }
    else {
      print PSOUT " ()";
      }
    }
  print PSOUT " ] def\n";

  # -- Directions 
  for ($d=0; $d<$ND; $d++) {
    $k = $d + $add;
    print PSOUT "/$DirArr[$d] [ ($Dir[$k])"; 
    for ($r=0; $r<$rounds; $r++) {
      print PSOUT " ($cell[$d][$r][$t])";
      if ($Rover>0 && $Rover==$cell[$d][$r][$t]) {
        $cell[$ND+1][$r][$t] = "-ROVER-";
        }
      }
    print PSOUT " ] def\n";
    }
 
  # -- Boards
  print PSOUT "/BD [ (Boards)";
  for ($r=0; $r<$rounds; $r++) {
    $k = $ND;
    $b1 = $cell[$k][$r][$t];    # apply board translation
    if ($b1 == 0) { $b2 = 'BYE'; }
    else {
      $b1 = $board[$b1-1];
      $b1 = ($b1*$BoardsPerRound)-$BoardsPerRound+1;
      if ($BoardsPerRound>1) {
        $b2 = $b1+$BoardsPerRound-1;
        $b2 = "$b1-$b2";
        }
      else {
        $b2 = $b1;
        }
      }
    print PSOUT " ($b2)";
    }
  print PSOUT " ] def\n";

  # -- Notes
  print PSOUT "/NT [ $Note";
  for ($r=0; $r<$rounds; $r++) {
    $k = $ND +1;
    print PSOUT " ($cell[$k][$r][$t])";
    }
  print PSOUT " ] def\n";

  print PSOUT <<"PrintEndMOV";
/NextEast   $NextEast def
/NextSouth  $NextSouth def
/NextWest   $NextWest def
/NextNorth  $NextNorth def
/TableNumber ($TNP[$p]) def
/Line1AdjNS $la1NS def
/Line1AdjEW $la1EW def
/Line2Adj   4 def
/Line2N     $L2N def
/Line2S     $L2S def
/Line2E     $L2E def
/Line2W     $L2W def
PrintEndMOV
  }

#----------------------------------------------------------------#

# -- Process File Name Argument

if ($#ARGV < 0) { Usage(); }

if ("$ARGV[0]" eq "-h") { Usage(); }
if ("$ARGV[0]" eq "-L") { License(); }

$file = $ARGV[$#ARGV];
pop @ARGV;
$FILE = $file;
$FILE =~ tr/a-z \t/A-Z/d; # uppercase without whitespace
print "File:     $FILE \n";

# -- Process the File Extension
if    ( $FILE =~ /.*MOV$/ ) { $TYPE = "MOV"; }
elsif ( $FILE =~ /.*IND$/ ) { $TYPE = "IND"; }
elsif ( $FILE =~ /.*BAM$/ ) { $TYPE = "BAM"; }
else  { die "--> file extension is not MOV, IND, or BAM: $FILE <--"; }

# -- Check File Existence
$BASEFILE = substr($FILE,0,length($FILE)-4);
$SWFILE = "BridgeMats.sw";
$PSin = "BridgeMats.ps";
( -e $FILE ) || 
    die "--> movement file does not exist: $FILE <--";
( -e $SWFILE ) || 
    warn "--> switch default file does not exist: $SWFILE <--";
( -e $PSin ) || 
    die "--> model PostScript file does not exist: $FILE <--";

# -- Initialize Command Line Switches
$DUMP=0;
$GROUP=1;
$FACT=1.0;
$BoardsPerRound=0;
$PrintLast=1;
$PrintMats=1;
$PrintOnly=0;
$Rover=0;
$NColor=1;
$TruncAfter=999;
$Weave=0;
for ($i=0; $i<10; $i++) { $desc[$i] = "";   $DI[$i] = 0; }

# -- Initialize Text Arrays
@Dir = ("N-S", "E-W", "N-S", "E-W", 
        "North", "East", "South", "West");
@Color = ("error","Black","Olive","Purple","Green","Blue","Teal","Red","Orange","Brown");
$MaxColor = 9;

# -- Begin Processing Files

ReadSwitchFile();
ProcessSwitches (1,@fsw);  # switch file
ProcessSwitches (2,@ARGV); # command line
$wide = int(400*$FACT+0.99);
$high = int(796*$FACT+0.99);

ReadMovementFile();

($PrintMats) || 
die "--> dump complete, not printing mats (-n) <--"; 
#($howell) || 
#die "--> BridgeMats is currently only for one-winner movements. <--"; 
($TYPE eq "BAM") &&
die "--> BridgeMats cannot do mats for Board-a-Match movements. <--";

# -- Open the PostScript input file
open (PSIN, $PSin) or 
    die "--> unable to open model PostScript file: $PSin <--\n";

# -- Open the PostScript output file
$PSFILE = ">$BASEFILE.PS";
open (PSOUT, $PSFILE) or 
    die "--> unable to open output file: \"$PSFILE\" <--\n";

# -- calculate number of pages
if    ( $PrintLast == 0 ) { $pages = $tables - 1; }
else                      { $pages = $tables; }
if    ( $PrintOnly > 0 )  {  $pages = 1; }

# -- get date info
@month = ('Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun',
          'Jul', 'Aug', 'Sep', 'Oct', 'Nov', 'Dec');
$sec=0; $wday=0; $yday=0; $tsdst=0; # (no -w msgs)
($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$tsdst) = localtime(time);
$year += 1900;

# -- read input header info and write it back out
$cut = 0;
while (<PSIN>) {
  if ( $_ =~ /^%ereh.tuc/ ) {
    $cut = 0;
    print PSOUT $_;
    }
  elsif ( $_ =~ /^%cut1here/ ) {
    $cut = 1;
    print PSOUT $_;
    PrintSetupVars;
    # -- process color request (after PrintSetupVars, changes players)
    if ($NColor==0) {
      if ($Rover==0 and ($pairs/2)==$tables and ($players/4)==$tables) {
        $NColor = ($tables-2)%($MaxColor-1) + 2;
        }
      else { $NColor = 1; }
      }
    ($DUMP) && print "Color:    $Color[$NColor]   ($NColor)\n";
    }
  elsif ( $_ =~ /^%%CreationDate:/ ) {
    print PSOUT "%%CreationDate: $mday-$month[$mon]-$year $hour:$min\n";
    }
  elsif ( $_ =~ /^%%Pages:/ ) {
    print PSOUT "%%Pages: $pages\n";
    }
  elsif ( $_ =~ /^%%BoundingBox:/ ) {
    print PSOUT "%%BoundingBox: 0 0 $wide $high\n";
    }
  elsif ( $_ =~ /^%%Page:/ ) {
    last;  # break out, once page data starts
    }
  else { 
    ($cut) || print PSOUT $_;
    } 
  }

# -- Read First Chunk of Page Data
$t = 0;
$cut = 0;
$Nchunk2 = 0;
while (<PSIN>) {
  if ( $_ =~ /^%cut2here/ ) {
    $cut = 2;
    $chunk1[$t] = $_;
    $Nchunk1 = $t + 1;
    }
  elsif ( $_ =~ /^%%PageBoundingBox:/ ) {
    $chunk1[$t] = "%%PageBoundingBox: 0 0 $wide $high\n";
    $Nchunk1 = $t + 1;
    }
  elsif ( $_ =~ /^%ereh.tuc/ ) {
    last;
    }
  else {
    ($cut) or $chunk1[$t] = $_;
    }
  $t++;
  }
($Nchunk1>0) || die "--> (Nchunk1=0) corrupt file: $PSin <--";
 
# -- read second chunk of page data
$chunk2[0] = $_;   # ereh2tuc
$t = 1;
$Nchunk2 = 0;
while (<PSIN>) {
  if ( $_ =~ /^%%EOF/ ) {
    $Nchunk2 = $t; # counts before current $t
    last;
    }
  else {
    $chunk2[$t] = $_;
    }
  $t++;
  }
($Nchunk2>0) || die "--> (Nchunk2=0) corrupt file: $PSin <--";

close (PSIN);

# -- Now Print the Pages
#        $p = page index
#        $t = table index
#        $r = round index
#        $d = direction index
#        $l = line index
#        $pn = page number
for ($p=0; $p<$pages; $p++) {
  $pn = $TN[$p];
  $t = $pn - 1;
  print PSOUT "%%Page: Table$TNP[$p]    $pn\n";
  for ($l=0; $l<$Nchunk1; $l++) {
    print PSOUT $chunk1[$l];
    }
  PrintPageVars();
  for ($l=0; $l<$Nchunk2; $l++) {
    print PSOUT $chunk2[$l];
    }
  }

print PSOUT "%%EOF\n";
close (PSOUT);
exit 0
